@@ -1,25 +1,25 @@
 """Python parser functions for EQL syntax."""
 from __future__ import unicode_literals
 
+import contextlib
 import datetime
 import re
 import sys
 from collections import defaultdict
-import contextlib
 
-from lark import Lark, Tree, Token
-from lark.visitors import Interpreter
+from lark import Lark, Token, Tree
 from lark.exceptions import LarkError
+from lark.visitors import Interpreter
 
-from . import ast
-from . import pipes
-from .types import TypeHint, NodeInfo, TypeFoldCheck
-from .errors import EqlSyntaxError, EqlSemanticError, EqlSchemaError, EqlTypeMismatchError, EqlError
+from . import ast, pipes
+from .errors import (EqlError, EqlSchemaError, EqlSemanticError,
+                     EqlSyntaxError, EqlTypeMismatchError)
 from .etc import get_etc_file
 from .functions import get_function, list_functions
 from .optimizer import Optimizer
 from .schema import EVENT_TYPE_ANY, EVENT_TYPE_GENERIC, Schema
-from .utils import to_unicode, load_extensions, ParserConfig, is_string
+from .types import NodeInfo, TypeFoldCheck, TypeHint
+from .utils import ParserConfig, is_string, load_extensions, to_unicode
 
 __all__ = (
     "allow_enum_fields",
@@ -43,7 +43,7 @@ __all__ = (
 )
 
 
-debugger_attached = 'pydevd' in sys.modules
+full_tracebacks = 'pydevd' in sys.modules
 
 NON_SPACE_WS = re.compile(r"[^\S ]+")
 
@@ -56,7 +56,8 @@ nullable_fields = ParserConfig(strict_fields=False)
 non_nullable_fields = ParserConfig(strict_fields=True)
 allow_enum_fields = ParserConfig(enable_enum=True)
 elasticsearch_syntax = ParserConfig(elasticsearch_syntax=True)
-
+elasticsearch_validate_optional_fields = ParserConfig(elasticsearch_syntax=True, validate_optional_fields=True)
+elastic_endpoint_syntax = ParserConfig(elasticsearch_syntax=True, dollar_var=True)
 
 keywords = ("and", "by", "const", "false", "in", "join", "macro",
             "not", "null", "of", "or", "sequence", "true", "until", "with", "where"
@@ -134,6 +135,8 @@ class LarkToEQL(Interpreter):
         self._ignore_missing = ParserConfig.read_stack("ignore_missing_fields", False)
         self._strict_fields = ParserConfig.read_stack("strict_fields", False)
         self._elasticsearch_syntax = ParserConfig.read_stack("elasticsearch_syntax", False)
+        self._dollar_var = ParserConfig.read_stack("dollar_var", False)
+        self._validate_optional_fields = ParserConfig.read_stack("validate_optional_fields", False)
         self._allow_enum = ParserConfig.read_stack("enable_enum", False)
         self._count_keys = []
         self._pipe_schemas = []
@@ -384,9 +387,9 @@ class LarkToEQL(Interpreter):
         return ast.TimeRange(quantity, unit)
 
     # fields
-    def _update_field_info(self, node_info):
+    def _update_field_info(self, node_info, optional_syntax=False):
         type_hint = None
-        allow_missing = self._schema.allow_missing
+        allow_missing = self._schema.allow_missing or (optional_syntax and not self._validate_optional_fields)
         field = node_info.node
         schema = None
         schema_hint = None
@@ -485,8 +488,12 @@ class LarkToEQL(Interpreter):
 
     def name(self, node):
         """Check for illegal use of keyword."""
-        text = node["NAME"].value
-        if text in keywords:
+        if isinstance(node, KvTree):
+            text = node.children[0].value
+        else:
+            text = str(node)
+
+        if text.rstrip("~") in keywords:
             raise self._error(node, "Invalid use of keyword", cls=EqlSyntaxError)
 
         return text
@@ -549,14 +556,33 @@ class LarkToEQL(Interpreter):
 
         return self._update_field_info(NodeInfo(ast.Field(name), source=node))
 
-    def field(self, node):
-        """Callback function to walk the AST."""
-        full_path = []
+    def varpath(self, node):
+        if not self._dollar_var:
+            raise self._error(node, "Invalid syntax", cls=EqlSyntaxError)
 
+        # no schema validation, just pass it through (syntax validation only)
+        if node["base_field"]:
+            path = [to_unicode(node["base_field"]["name"])]
+        else:
+            _, path = self._field_path(node["field"])
+
+        field = ast.Field(path[0], path[1:], as_var=True)
+        return NodeInfo(field, source=node, type_info=TypeHint.Unknown)
+
+    def _field_path(self, node, allow_optional=False):
+        full_path = []
         # to get around parser ambiguities, we had to create a token to mash all of the parts together
         # but we have a separate rule "field_parts" that can safely re-parse and separate out the tokens.
         # we can walk through each token, and build the field path accordingly
-        for part in lark_parser.parse(node.children[0], "field_parts").children:
+        text = node.children[-1]
+        optional_syntax = text.startswith("?")
+        if optional_syntax:
+            if not allow_optional:
+                raise self._error(node, "Optional fields are not supported.", cls=EqlSyntaxError, width=1)
+
+            text = text[1:]
+
+        for part in lark_parser.parse(text, "field_parts").children:
             if part["NAME"]:
                 name = to_unicode(part["NAME"])
                 full_path.append(name)
@@ -568,8 +594,13 @@ class LarkToEQL(Interpreter):
             elif part["UNSIGNED_INTEGER"]:
                 full_path.append(int(part["UNSIGNED_INTEGER"]))
             else:
-                raise self._error(node, "Unable to parser field", cls=EqlSyntaxError)
+                raise self._error(node, "Unable to parse field", cls=EqlSyntaxError)
 
+        return optional_syntax, full_path
+
+    def field(self, node):
+        """Callback function to walk the AST."""
+        optional_syntax, full_path = self._field_path(node, allow_optional=self._elasticsearch_syntax)
         base, path = full_path[0], full_path[1:]
 
         # if get_variable:
@@ -581,12 +612,24 @@ class LarkToEQL(Interpreter):
         #     # This can be overridden by the parent function that is parsing it
         #     return self._add_variable(node.base)
         field = ast.Field(base, path)
-        return self._update_field_info(NodeInfo(field, source=node))
+        return self._update_field_info(NodeInfo(field, source=node), optional_syntax=optional_syntax)
 
-    def ilike(self, node):
+    def string_predicate(self, node):
         """Callback function to walk the AST."""
+        predicate = node["STRING_PREDICATE"]
+        if predicate == ":":
+            function_name = "wildcard"
+        elif predicate in ("like", "like~"):
+            function_name = "wildcard"
+        elif predicate in ("regex", "regex~"):
+            function_name = "match"
+        else:
+            raise self._error(node, message="Invalid syntax", cls=EqlSyntaxError)
+
         if not self._elasticsearch_syntax:
-            raise self._error(node, message="Invalid syntax, try == or wildcard()", cls=EqlSyntaxError)
+            args = ", ".join(self.text[n.meta.start_pos:n.meta.end_pos] for n in node.child_trees)
+            raise self._error(node, "Invalid syntax. Try: {function_name}({args})", cls=EqlSyntaxError,
+                              function_name=function_name, args=args)
 
         children = self.visit(node.child_trees)
 
@@ -594,7 +637,7 @@ class LarkToEQL(Interpreter):
             if not child.validate_type(TypeHint.String):
                 raise self._type_error(child, TypeHint.String)
 
-        return NodeInfo(ast.FunctionCall("wildcard", [child.node for child in children], TypeHint.Boolean))
+        return NodeInfo(ast.FunctionCall(function_name, [child.node for child in children]), TypeHint.Boolean)
 
     def comparison(self, node):
         """Callback function to walk the AST."""
@@ -641,21 +684,22 @@ class LarkToEQL(Interpreter):
         eql_node = ast.Comparison(left.node, op, right.node)
 
         # there is no special comparison operator for wildcards, just look for * in the string
-        if not self._elasticsearch_syntax and isinstance(right.node, ast.String) and '*' in right.node.value:
-            func_call = ast.FunctionCall('wildcard', [left.node, right.node])
+        if not self._elasticsearch_syntax:
+            if isinstance(right.node, ast.String) and '*' in right.node.value:
+                func_call = ast.FunctionCall('wildcard', [left.node, right.node])
 
-            if op == ast.Comparison.EQ:
-                eql_node = func_call
-            elif op == ast.Comparison.NE:
-                eql_node = ~func_call
+                if op == ast.Comparison.EQ:
+                    eql_node = func_call
+                elif op == ast.Comparison.NE:
+                    eql_node = ~func_call
 
-        elif not self._elasticsearch_syntax and isinstance(left.node, ast.String) and '*' in left.node.value:
-            func_call = ast.FunctionCall('wildcard', [right.node, left.node])
+            elif isinstance(left.node, ast.String) and '*' in left.node.value:
+                func_call = ast.FunctionCall('wildcard', [right.node, left.node])
 
-            if op == ast.Comparison.EQ:
-                eql_node = func_call
-            elif op == ast.Comparison.NE:
-                eql_node = ~func_call
+                if op == ast.Comparison.EQ:
+                    eql_node = func_call
+                elif op == ast.Comparison.NE:
+                    eql_node = ~func_call
 
         return NodeInfo(eql_node, TypeHint.Boolean, nullable=left.nullable or right.nullable, source=node)
 
@@ -727,7 +771,11 @@ class LarkToEQL(Interpreter):
 
     def in_set(self, node):
         """Callback function to walk the AST."""
-        outer, container = self.visit_children(node)  # type: (NodeInfo, list[NodeInfo])
+        if not self._elasticsearch_syntax and node.get("IN") == "in~":
+            raise self._error(node, message="Invalid syntax. Explicit case-insensitivity is not supported.",
+                              cls=EqlSyntaxError)
+
+        outer, container = self.visit(node.child_trees)  # type: (NodeInfo, list[NodeInfo])
 
         if not outer.validate_type(TypeHint.primitives()):
             # can't compare non-primitives to sets
@@ -770,10 +818,24 @@ class LarkToEQL(Interpreter):
 
     def function_call(self, node, prev_node=None, prev_arg=None):
         """Callback function to walk the AST."""
-        function_name = self.visit(node["name"] or node["method_name"])
+        if node.data == "method":
+            name_node = node["method_name"]
+            function_name = self.method_name(name_node)
+        else:
+            name_node = node.children[0]
+            function_name = self.name(name_node)
+
         args = []
         nodes = []
 
+        if function_name.endswith("~"):
+            # remove the trailing ~ and use the existing AST
+            function_name = function_name.rstrip("~")
+
+            if not self._elasticsearch_syntax:
+                raise self._error(node, "Invalid syntax. Explicit case-insensitivity is not supported",
+                                  cls=EqlSyntaxError)
+
         non_method_arguments = node["expressions"].children if node["expressions"] else []
 
         if prev_node:
@@ -840,7 +902,7 @@ class LarkToEQL(Interpreter):
             return NodeInfo(expr, signature.return_value, nullable=nullable, source=node)
 
         elif self._check_functions:
-            raise self._error(node["name"], "Unknown function {NAME}")
+            raise self._error(name_node, "Unknown function {function_name}", function_name=function_name)
         else:
             args = []
 
@@ -940,7 +1002,7 @@ class LarkToEQL(Interpreter):
         query = self.visit(node["subquery"]["event_query"])
         return NodeInfo(ast.NamedSubquery(name, query), TypeHint.Boolean, source=node)
 
-    def subquery_by(self, node, num_values=None, position=None, close=None, allow_fork=False):
+    def subquery_by(self, node, num_values=None, position=None, close=None, allow_fork=False, allow_runs=False):
         """Callback function to walk the AST."""
         if not self._subqueries_enabled:
             raise self._error(node, "Subqueries not supported")
@@ -961,6 +1023,20 @@ class LarkToEQL(Interpreter):
             raise self._error(error_node, message, num=num_values, end=end)
 
         kwargs = {}
+
+        repeated_sequence = node["repeated_sequence"]
+
+        runs_count = 1
+        if repeated_sequence is not None:
+            runs_count = int(node["repeated_sequence"]["UNSIGNED_INTEGER"].value)
+
+            if allow_runs is False:
+                raise self._error(repeated_sequence, "Unsupported usage of repeated syntax", cls=EqlSyntaxError)
+
+            if runs_count <= 1:
+                raise self._error(repeated_sequence, "Repeated sequence runs must be greater than 1",
+                                  cls=EqlSemanticError)
+
         fork_param = node["fork_param"]
 
         if fork_param is not None:
@@ -978,7 +1054,8 @@ class LarkToEQL(Interpreter):
         else:
             join_values = []
 
-        return NodeInfo(ast.SubqueryBy(query, [v.node for v in join_values], **kwargs), source=node), join_values
+        node_info = NodeInfo(ast.SubqueryBy(query, [v.node for v in join_values], **kwargs), source=node)
+        return node_info, join_values, runs_count
 
     def join_values(self, node):
         """Return all of the expressions."""
@@ -989,7 +1066,7 @@ class LarkToEQL(Interpreter):
         queries, close = self._get_subqueries_and_close(node)
         return ast.Join(queries, close)
 
-    def _get_subqueries_and_close(self, node, allow_fork=False):
+    def _get_subqueries_and_close(self, node, allow_fork=False, allow_runs=False):
         """Helper function used by join and sequence to avoid duplicate code."""
         if not self._subqueries_enabled:
             # Raise the error earlier (instead of waiting until subquery_by) so that it's more meaningful
@@ -997,7 +1074,8 @@ class LarkToEQL(Interpreter):
 
         # Figure out how many fields are joined by in the first query, and match across all
         subquery_nodes = node.get_list("subquery_by")
-        first, first_info = self.subquery_by(subquery_nodes[0], allow_fork=allow_fork, position=0)
+        first, first_info, _ = self.subquery_by(subquery_nodes[0], allow_fork=allow_fork,
+                                                position=0, allow_runs=allow_runs)
         num_values = len(first_info)
         subqueries = [(first, first_info)]
 
@@ -1006,10 +1084,16 @@ class LarkToEQL(Interpreter):
         close = None
 
         if until_node:
+            repeated_sequence = until_node["subquery_by"]["repeated_sequence"]
+            if repeated_sequence:
+                raise self._error(repeated_sequence, "Unsupported usage of repeated syntax", cls=EqlSyntaxError)
             subquery_nodes.append(until_node["subquery_by"])
 
         for pos, subquery in enumerate(subquery_nodes[1:], 1):
-            subqueries.append(self.subquery_by(subquery, num_values=num_values, allow_fork=allow_fork, position=pos))
+            subquery, join_values, runs_count = self.subquery_by(subquery, num_values=num_values, allow_fork=allow_fork,
+                                                                 position=pos, allow_runs=allow_runs)
+            multiple_subqueries = [(subquery, join_values)] * runs_count
+            subqueries.extend(multiple_subqueries)
 
         # Validate that each field has matching types
         default_hint = TypeHint.primitives()
@@ -1100,7 +1184,12 @@ class LarkToEQL(Interpreter):
         if node['with_params']:
             params = self.time_range(node['with_params']['time_range'])
 
-        queries, close = self._get_subqueries_and_close(node, allow_fork=True)
+        allow_runs = self._elasticsearch_syntax
+
+        queries, close = self._get_subqueries_and_close(node, allow_fork=True, allow_runs=allow_runs)
+        if len(queries) <= 1 and not self._elasticsearch_syntax:
+            raise self._error(node, "Only one item in the sequence",
+                              cls=EqlSemanticError if self._elasticsearch_syntax else EqlSyntaxError)
         return ast.Sequence(queries, params, close)
 
     def definitions(self, node):
@@ -1169,7 +1258,7 @@ def _parse(text, start=None, preprocessor=None, implied_any=False, implied_base=
             exc = EqlSyntaxError("Invalid syntax",
                                  line=e.line - 1, column=e.column - 1, width=width,
                                  source='\n'.join(walker.lines[e.line - 2:e.line]))
-            if cfg.read_stack("full_traceback", debugger_attached):
+            if cfg.read_stack("full_traceback", full_tracebacks):
                 raise exc
 
         if exc is None:
@@ -1184,7 +1273,7 @@ def _parse(text, start=None, preprocessor=None, implied_any=False, implied_base=
                 return eql_node
             except EqlError as e:
                 # If full traceback mode is enabled, then re-raise the exception
-                if cfg.read_stack("full_traceback", debugger_attached):
+                if cfg.read_stack("full_traceback", full_tracebacks):
                     raise
                 exc = e
 
